# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Search tools.
"""

from itertools import chain
import json

import numpy as np
import pandas as pd
from tornado.template import Template

from nlg import utils
from nlg import grammar

SEARCH_PRIORITIES = [
    {'type': 'ne'},  # A match which is a named entity gets the highest priority
    {'location': 'fh_args'},  # than one that is a formhandler arg
    {'location': 'colname'},  # than one that is a column name
    {'type': 'quant'},  # etc
    {'location': 'cell'}
]


def _sort_search_results(items, priorities=SEARCH_PRIORITIES):
    """
    Sort a list of search results by `priorities`.

    Parameters
    ----------
    items : dict
        Dictionary containing search results, where keys are tokens and values
        are lists of locations where the token was found. Preferably this should
        be a `DFSearchResults` object.
    priorities : list, optional
        List of rules that allow sorting of search results. A `rule` is any
        subset of a search result dictionary. Lower indices indicate higher priorities.

    Returns
    -------
    dict
        Prioritized search results - for each {token: search_matches} pair, sort
        search_matches such that a higher priority search result is enabled.
    """
    match_ix = [[p.items() <= item.items() for p in priorities] for item in items]
    min_match = [m.index(True) for m in match_ix]
    items[min_match.index(min(min_match))]['enabled'] = True
    return items


class DFSearchResults(dict):
    """A convenience wrapper around `dict` to collect search results.

    Different from `dict` in that values are always lists, and setting to
    existing key appends to the list."""

    def __setitem__(self, key, value):
        if key not in self:
            super(DFSearchResults, self).__setitem__(key, [value])
        elif self[key][0] != value:
            self[key].append(value)

    def update(self, other):
        # Needed because the default update method doesn't seem to use setitem
        for k, v in other.items():
            self[k] = v

    def clean(self):
        """Sort the search results for each token by priority and un-overlap tokens."""
        for k, v in self.items():
            _sort_search_results(v)
        # unoverlap the keys
        to_remove = []
        for k in self:
            if any([k in c for c in self.keys() - {k}]):
                to_remove.append(k)
        for i in to_remove:
            del self[i]


class DFSearch(object):
    """Make a dataframe searchable."""

    def __init__(self, df, nlp=None, **kwargs):
        """Default constrictor.

        Parameters
        ----------
        df : pd.DataFrame
            The dataframe to search.
        nlp : A `spacy.lang` model, optional
        """
        self.df = df
        # What do results contain?
        # A map of tokens to list of search results.
        self.results = DFSearchResults()
        if not nlp:
            nlp = utils.load_spacy_model()
        self.nlp = nlp
        self.matcher = kwargs.get('matcher', utils.make_np_matcher(self.nlp))

    def search(self, text, colname_fmt='df.columns[{}]',
               cell_fmt='df["{}"].iloc[{}]', **kwargs):
        """
        Search the dataframe.

        Parameters
        ----------
        text : str
            The text to search.
        colname_fmt : str, optional
            String format to describe dataframe columns in the search results,
            can be one of 'df.columns[{}]' or 'df[{}]'.
        cell_fmt : str, optional
            String format to describe dataframe values in the search results.
            Can be one of 'df.iloc[{}, {}]', 'df.loc[{}, {}]', 'df[{}][{}]', etc.

        Returns
        -------
        dict
            A dictionary who's keys are tokens from `text` found in
            the source dataframe, and values are a list of locations in the df
            where they are found.
        """
        self.search_nes(text)
        for token, ix in self.search_columns(text, **kwargs).items():
            ix = utils.sanitize_indices(self.df.shape, ix, 1)
            self.results[token] = {'location': 'colname', 'tmpl': colname_fmt.format(ix),
                                   'type': 'token'}

        for token, (x, y) in self.search_table(text, **kwargs).items():
            x = utils.sanitize_indices(self.df.shape, x, 0)
            y = utils.sanitize_indices(self.df.shape, y, 1)
            self.results[token] = {
                'location': 'cell', 'tmpl': cell_fmt.format(self.df.columns[y], x),
                'type': 'token'}
        self.search_quant([c.text for c in self.doc if c.pos_ == 'NUM'])
        return self.results

    def search_nes(self, text, colname_fmt='df.columns[{}]', cell_fmt='df["{}"].iloc[{}]'):
        """Find named entities in text, and search for them in the dataframe.

        Parameters
        ----------
        text : str
            The text to search.
        """
        self.doc = self.nlp(text)
        self.ents = utils.ner(self.doc, self.matcher)
        ents = [c.text for c in self.ents]
        for token, ix in self.search_columns(ents, literal=True).items():
            ix = utils.sanitize_indices(self.df.shape, ix, 1)
            self.results[token] = {
                'location': 'colname',
                'tmpl': colname_fmt.format(ix), 'type': 'ne'
            }
        for token, (x, y) in self.search_table(ents, literal=True).items():
            x = utils.sanitize_indices(self.df.shape, x, 0)
            y = utils.sanitize_indices(self.df.shape, y, 1)
            self.results[token] = {
                'location': 'cell',
                'tmpl': cell_fmt.format(self.df.columns[y], x), 'type': 'ne'}

    def search_table(self, text, **kwargs):
        """Search the `.values` attribute of the dataframe for tokens in `text`."""
        kwargs['array'] = self.df.copy()
        return self._search_array(text, **kwargs)

    def search_columns(self, text, **kwargs):
        """Search df columns for tokens in `text`."""
        kwargs['array'] = self.df.columns
        return self._search_array(text, **kwargs)

    def search_quant(self, quants, nround=2, cell_fmt='df["{}"].iloc[{}]'):
        """Search the dataframe for a set of quantitative values.

        Parameters
        ----------
        quants : list / array like
            The values to search.
        nround : int, optional
            Numeric values in the dataframe are rounded to these many
            significant digits before searching.
        """
        dfclean = utils.sanitize_df(self.df, nround)
        quants = np.array(quants)
        n_quant = quants.astype('float').round(2)
        for x, y in zip(*dfclean.isin(n_quant).values.nonzero()):
            x = utils.sanitize_indices(dfclean.shape, x, 0)
            y = utils.sanitize_indices(dfclean.shape, y, 1)
            tk = quants[n_quant == dfclean.iloc[x, y]][0].item()
            self.results[tk] = {
                'location': 'cell', 'tmpl': cell_fmt.format(self.df.columns[y], x),
                'type': 'quant'}

    def _search_array(self, text, array, literal=False,
                      case=False, lemmatize=True, nround=2):
        """Search for tokens in text within an array.

        Parameters
        ----------
        text : str or spacy document
            Text to search
        array : array-like
            Array to search in.
        literal : bool, optional
            Whether to match tokens to values literally.
        case : bool, optional
            If true, run a case sensitive search.
        lemmatize : bool, optional
            If true (default), search on lemmas of tokens and values.
        nround : int, optional
            Significant digits used to round `array` before searching.

        Returns
        -------
        dict
            Mapping of tokens to a sequence of indices within `array`.

        Example
        -------
        >>> _search_array('3', np.arange(5))
        {'3': [2]}
        >>> df = pd.DataFrame(np.eye(3), columns='one punch man'.split())
        >>> _search_array('1', df.values)
        {'1': [(0, 0), (1, 1), (2, 2)]}
        >>> _search_array('punched man', df.columns)
        {'punched': [1], 'man': [2]}
        >>> _search_array('1 2 buckle my shoe', df.index)
        {'1': [1], '2': [2]}
        """
        if literal:
            # Expect text to be a list of strings, no preprocessing on anything.
            if not isinstance(text, list):
                raise TypeError('text is expected to be list of strs when literal=True.')
            valid_types = {float, int, str}
            if not set([type(c) for c in text]).issubset(valid_types):
                raise TypeError('text can contain only strings or numbers when literal=True.')
            tokens = {c: str(c) for c in text}
        elif lemmatize:
            tokens = {c.lemma_: c.text for c in self.nlp(text)}
            if array.ndim == 1:
                array = [c if isinstance(c, str) else str(c) for c in array]
                array = [self.nlp(c) for c in array]
                array = pd.Series([token.lemma_ for doc in array for token in doc])
            else:
                for col in array.columns[array.dtypes == np.dtype('O')]:
                    s = [c if isinstance(c, str) else str(c) for c in array[col]]
                    s = [self.nlp(c) for c in s]
                    try:
                        array[col] = [token.lemma_ for doc in s for token in doc]
                    except ValueError:
                        # You cannot lemmatize columns that have multi-word values
                        if not case:  # still need to respect the `case` param
                            array[col] = array[col].str.lower()
        else:
            if not case:
                tokens = {c.text.lower(): c.text for c in self.nlp(text)}
                if array.ndim == 1:
                    array = array.str.lower()
                else:
                    for col in array.columns[array.dtypes == np.dtype('O')]:
                        array[col] = array[col].str.lower()
            else:
                tokens = {c.text: c.text for c in self.nlp(text)}
        mask = array.isin(tokens.keys())
        if mask.ndim == 1:
            if mask.any():
                ix = mask.nonzero()[0]
                return {tokens[array[i]]: i for i in ix}
            return {}
        else:
            if mask.any().any():
                ix, iy = mask.values.nonzero()
                return {tokens[array.iloc[x, y]]: (x, y) for x, y in zip(ix, iy)}
        return {}


def search_args(entities, args, lemmatized=True, fmt='fh_args["{}"][{}]',
                argkeys=('_sort', '_by', '_c')):
    """
    Search formhandler arguments provided as URL query parameters.

    Parameters
    ----------
    entities : list
        list of named entities found in the source text
    args : dict
        FormHandler args as parsed by g1.url.parse(...).searchList
    lemmatized : bool, optional
        whether to search on lemmas of text values
    fmt : str, optional
        String format used to describe FormHandler arguments in the template
    argkeys : list, optional
        Formhandler argument keys to be considered for the search. Any key not
        present in this will be ignored.
        # TODO: Column names can be keys too!!

    Returns
    -------
    dict
        Mapping of entities / tokens to objects describing where they are found
        in Formhandler arguemnts. Each search result object has the following
        structure:
        {
            'type': 'some token',
            'location': 'fh_args',
            'tmpl': 'fh_args['_by'][0]'  # The template that gets this token from fh_args
        }
    """
    nlp = utils.load_spacy_model()
    args = {k: v for k, v in args.items() if k in argkeys}
    search_res = {}
    ent_tokens = list(chain(*entities))
    for k, v in args.items():
        v = [t.lstrip('-') for t in v]
        # argtokens = list(chain(*[re.findall(r'\w+', f) for f in v]))
        argtokens = list(chain(*[nlp(c) for c in v]))
        for i, x in enumerate(argtokens):
            for y in ent_tokens:
                if lemmatized:
                    if x.lemma_ == y.lemma_:
                        search_res[y.text] = {
                            'type': 'token', 'tmpl': fmt.format(k, i),
                            'location': 'fh_args'}
                else:
                    if x.text == y.text:
                        search_res[y.text] = {
                            'type': 'token', 'tmpl': fmt.format(k, i),
                            'location': 'fh_args'}
    return search_res


def _search(text, args, df):
    """Construct a tornado template which regenerates some
    text from a dataframe and formhandler arguments.

    The pipeline consists of:
    1. cleaning the text and the dataframe
    2. searching the dataframe and FH args for tokens in the text
    3. detecting inflections on the tokens.

    Parameters
    ----------
    text : str
        Input text
    args : dict
        Formhandler arguments
    df : pd.DataFrame
        Source dataframe.

    Returns
    --------
    tuple
        of search results, cleaned text and token inflections. The webapp uses
        these to construct a tornado template.
    """
    utils.load_spacy_model()
    df = utils.grmfilter(df, args.copy())
    clean_text = utils.sanitize_text(text)
    args = utils.sanitize_fh_args(args)
    # Is this correct?
    dfs = DFSearch(df)
    dfix = dfs.search(clean_text)
    dfix.update(search_args(dfs.ents, args))
    dfix.clean()
    inflections = grammar.find_inflections(clean_text, dfix, args, df)
    _infl = {}
    for token, funcs in inflections.items():
        _infl[token] = []
        for func in funcs:
            _infl[token].append({
                'source': func.source,
                'fe_name': func.fe_name,
                'func_name': func.__name__
            })
    return dfix, clean_text, _infl


def _make_inflection_string(tmpl, infl):
    source = infl['source']
    func_name = infl['func_name']
    if source == 'str':
        tmpl += f'.{func_name}()'
    else:
        tmpl = f'{source}.{func_name}({tmpl})'
    return tmpl


t_templatize = lambda x: '{{ ' + x + ' }}'  # noqa: E731


def templatize_token(token, results, inflection):
    for r in results:
        if r.get('enabled', False):
            break
    tmpl = r['tmpl']
    if inflection:
        for i in inflection:
            tmpl = _make_inflection_string(tmpl, i)
    return t_templatize(tmpl)


def set_fh_args(text, fh_args):
    fh_args = json.dumps(fh_args)
    tmpl = f'{{% set fh_args = {fh_args}  %}}\n'
    tmpl += f'{{% set df = U.grmfilter(orgdf, fh_args.copy()) %}}\n'
    return tmpl + text


def templatize(text, args, df):
    """Construct a tornado template which regenerates some
    text from a dataframe and formhandler arguments.

    Parameters
    ----------
    text : str
        Input text
    args : dict
        Formhandler arguments
    df : pd.DataFrame
        Source dataframe.

    Returns
    -------
    str
        Tornado template corresponding to the text and data.

    Example
    -------
    >>> from gramex import data
    >>> df = pd.read_csv('iris.csv')
    >>> fh_args = {'_by': ['species']}
    >>> df = data.filter(df, fh_args.copy())
    >>> text = 'The iris dataset has 3 species - setosa, versicolor and virginica.'
    >>> template = templatize(text, fh_args, df)
    >>> print(template)
    {% set fh_args = {"_by": ["species"]}  %}
    {% set df = U.grmfilter(orgdf, fh_args.copy()) %}
    The iris dataset has 3 {{ df.columns[0] }} - {{ df["species"].iloc[0] }}, \
{{ df["species"].iloc[1] }} and {{ df["species"].iloc[-1] }}.
    """
    dfix, clean_text, infl = _search(text, args, df)
    sentence = text
    for tk, tkobj in dfix.items():
        inflection = infl.get(tk)
        tk_tmpl = templatize_token(tk, tkobj, inflection)
        sentence = sentence.replace(tk, tk_tmpl)
    return set_fh_args(sentence, args)


def render(df, template):
    return Template(template).generate(orgdf=df, U=utils, G=grammar)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""WIP: The Narrative class."""


class Narrative(object):
    def __init__(self, sentences=None, conditions=None):
        if sentences is None:
            sentences = []
        self.sentences = sentences
        if conditions is None:
            conditions = []
        self.conditions = conditions

    @property
    def sdict(self):
        return {k: i for i, k in enumerate(self.sentences)}

    def templatize(self, sep="\n\n"):
        newsents = []
        for i, sent in enumerate(self.sentences):
            if str(i) in self.conditions:
                newsent = """
                {{ % if df.eval(\'{expr}\').any() % }}
                    {sent}
                {{ % end % }}
                """.format(
                    expr=self.conditions[str(i)], sent=sent
                )
            else:
                newsent = sent
            newsents.append(newsent)
        return sep.join(newsents)

    def append(self, sent):
        self.sentences.append(sent)

    add = append

    def prepend(self, sent):
        self.sentences.insert(0, sent)

    def insert(self, sent, ix):
        self.sentences.insert(ix, sent)

    def _move(self, key, pos):
        if isinstance(key, str):
            key = self.sdict[key]
        self.sentences.insert(key + pos, self.sentences.pop(key))

    def move_up(self, key):
        self._move(key, 1)

    def move_down(self, key):
        self._move(key, -1)



import sys

__version__ = '0.1'

try:
    __NLG_SETUP__
except NameError:
    __NLG_SETUP__ = False


if __NLG_SETUP__:
    sys.stderr.write('Partial import of nlg during the build process.\n')
else:
    from .search import templatize  # NOQA: F401
    from .narrative import Narrative  # NOQA: F401
    __all__ = ['templatize', 'Narrative']



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Tests of the nlg.search module
"""

import os.path as op
import re
import unittest

import pandas as pd
from tornado.template import Template

from nlg import search, utils

nlp = utils.load_spacy_model()
matcher = utils.make_np_matcher(nlp)


class TestDFSearch(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        fpath = op.join(op.dirname(__file__), "data", "actors.csv")
        cls.df = pd.read_csv(fpath, encoding='utf-8')
        cls.dfs = search.DFSearch(cls.df)

    def test__search_array(self):
        sent = "The votes, names and ratings of artists."
        res = self.dfs._search_array(sent, self.df.columns, lemmatize=False)
        self.assertDictEqual(res, {'votes': 3})

        res = self.dfs._search_array(sent, self.df.columns)
        self.assertDictEqual(res, {'votes': 3, 'names': 1, 'ratings': 2})

        sent = "The votes, NAME and ratings of artists."
        res = self.dfs._search_array(sent, self.df.columns,
                                     lemmatize=False)
        self.assertDictEqual(res, {'votes': 3, 'NAME': 1})
        res = self.dfs._search_array(sent, self.df.columns, lemmatize=False,
                                     case=True)
        self.assertDictEqual(res, {'votes': 3})

    def test_dfsearch_lemmatized(self):
        df = pd.DataFrame.from_dict(
            {
                "partner": ["Lata Mangeshkar", "Asha Bhosale", "Mohammad Rafi"],
                "song": [20, 5, 15],
            }
        )
        sent = "Kishore Kumar sang the most songs with Lata Mangeshkar."
        dfs = search.DFSearch(df)
        self.assertDictEqual(
            dfs.search(sent, lemmatize=True),
            {
                'songs': [{"location": "colname", "type": "token", "tmpl": "df.columns[1]"}],
                'Lata Mangeshkar': [
                    {'location': 'cell', 'tmpl': 'df["partner"].iloc[0]', 'type': 'ne'}],
            }
        )

    def test_search_df(self):
        fpath = op.join(op.dirname(__file__), "data", "actors.csv")
        df = pd.read_csv(fpath, encoding='utf-8')
        df.sort_values("votes", ascending=False, inplace=True)
        df.reset_index(inplace=True, drop=True)
        dfs = search.DFSearch(df)
        sent = "Spencer Tracy is the top voted actor."
        self.assertDictEqual(
            dfs.search(sent),
            {
                'Spencer Tracy': [
                    {'location': 'cell', 'tmpl': 'df["name"].iloc[0]', 'type': 'ne'}
                ],
                'voted': [{'location': 'colname', 'tmpl': 'df.columns[-1]', 'type': 'token'}],
                'actor': [{'location': 'cell', 'tmpl': 'df["category"].iloc[-4]', 'type': 'token'}]
            }
        )


class TestSearch(unittest.TestCase):
    def test_dfsearches(self):
        x = search.DFSearchResults()
        x['hello'] = 'world'
        x['hello'] = 'world'
        self.assertDictEqual(x, {'hello': ['world']})
        x = search.DFSearchResults()
        x['hello'] = 'world'
        x['hello'] = 'underworld'
        self.assertDictEqual(x, {'hello': ['world', 'underworld']})

    @unittest.skip("Temporary")
    def test_search_args(self):
        args = {"_sort": ["-votes"]}
        doc = nlp("James Stewart is the top voted actor.")
        ents = utils.ner(doc, matcher)
        self.assertDictEqual(
            search.search_args(ents, args),
            {
                "voted": {
                    "tmpl": "fh_args['_sort'][0]",
                    "type": "token",
                    "location": "fh_args"
                }
            }
        )

    @unittest.skip("Temporary")
    def test_search_args_literal(self):
        args = {"_sort": ["-rating"]}
        doc = nlp("James Stewart has the highest rating.")
        ents = utils.ner(doc, matcher)
        self.assertDictEqual(search.search_args(ents, args, lemmatized=False),
                             {'rating': {
                                 "tmpl": "fh_args['_sort'][0]",
                                 "location": "fh_args",
                                 "type": "token"}})

    def test_templatize(self):
        fpath = op.join(op.dirname(__file__), "data", "actors.csv")
        df = pd.read_csv(fpath, encoding='utf-8')
        df.sort_values("votes", ascending=False, inplace=True)
        df.reset_index(inplace=True, drop=True)

        doc = """
        Spencer Tracy is the top voted actor, followed by Cary Grant.
        The least voted actress is Bette Davis, trailing at only 14 votes, followed by
        Ingrid Bergman at a rating of 0.29614.
        """
        ideal = """
        {{ df['name'].iloc[0] }} is the top {{ fh_args['_sort'][0] }}
        {{ df['category'].iloc[-4] }}, followed by {{ df['name'].iloc[1] }}.
        The least {{ fh_args['_sort'][0] }} {{ df['category'].iloc[-1] }} is
        {{ df['name'].iloc[-1] }}, trailing at only {{ df['votes'].iloc[-1] }}
        {{ df.columns[-1] }}, followed by {{ df['name'].iloc[-2] }} at a {{ df.columns[2] }}
        of {{ df['rating'].iloc[-2] }}.
        """
        args = {"_sort": ["-votes"]}
        tokenmap, text, inflections = search._search(doc, args, df)
        actual = text
        for token, tmpls in tokenmap.items():
            tmpl = [t for t in tmpls if t.get('enabled', False)][0]
            actual = actual.replace(token, '{{{{ {} }}}}'.format(tmpl['tmpl']))
        cleaner = lambda x: re.sub(r"\s+", " ", x)  # NOQA: E731
        ideal, actual = map(cleaner, (ideal, actual))
        ideal = Template(ideal).generate(df=df, fh_args=args)
        actual = Template(actual).generate(df=df, fh_args=args)
        self.assertEqual(ideal, actual)
        self.assertDictEqual(
            inflections,
            {
                'actor': [{'fe_name': 'Singularize', 'source': 'G', 'func_name': 'singular'}],
                'actress': [{'source': 'G', 'fe_name': 'Singularize', 'func_name': 'singular'}]
            }
            # Don't detect inflections until they can be processed without intervention
            # 'voted': [{'source': 'G', 'fe_name': 'Lemmatize', 'func_name': 'lemmatize'}]}
        )

    def test_search_sort(self):

        results = [
            {'tmpl': 'df.loc[0, "name"]', 'type': 'ne', 'location': 'cell'},
            {'tmpl': 'df.columns[0]', 'type': 'token', 'location': 'colname'},
            {'tmpl': 'args["_sort"][0]', 'type': 'token', 'location': 'fh_args'}
        ]
        _sorted = search._sort_search_results(results)
        enabled = [c for c in _sorted if c.get('enabled', False)]
        self.assertListEqual(enabled, results[:1])

        results = [
            {'tmpl': 'df.columns[0]', 'type': 'token', 'location': 'colname'},
            {'tmpl': 'args["_sort"][0]', 'type': 'token', 'location': 'fh_args'},
            {'tmpl': 'df["foo"].iloc[0]', 'type': 'token', 'location': 'cell'}
        ]
        _sorted = search._sort_search_results(results)
        enabled = [c for c in _sorted if c.get('enabled', False)]
        self.assertListEqual(enabled, results[1:2])

        results = [
            {'tmpl': 'df.columns[0]', 'type': 'token', 'location': 'colname'},
            {'tmpl': 'args["_sort"][0]', 'type': 'token', 'location': 'cell'},
            {'tmpl': 'df["foo"].iloc[0]', 'type': 'quant', 'location': 'cell'}
        ]
        _sorted = search._sort_search_results(results)
        enabled = [c for c in _sorted if c.get('enabled', False)]
        self.assertListEqual(enabled, results[:1])

        results = [
            {'tmpl': 'args["_sort"][0]', 'type': 'token', 'location': 'cell'},
            {'tmpl': 'df["foo"].iloc[0]', 'type': 'quant', 'location': 'cell'}
        ]
        _sorted = search._sort_search_results(results)
        enabled = [c for c in _sorted if c.get('enabled', False)]
        self.assertListEqual(enabled, results[1:])


if __name__ == "__main__":
    unittest.main()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""Tests for nlg.utils"""

import unittest
from nlg import utils


nlp = utils.load_spacy_model()
matcher = utils.make_np_matcher(nlp)


class TestUtils(unittest.TestCase):

    def test_join_words(self):
        sent = 'The quick brown fox jumps over the lazy dog.'
        self.assertEqual(utils.join_words(sent), sent.rstrip('.'))
        self.assertEqual(utils.join_words(sent, ''), sent.rstrip('.').replace(' ', ''))
        self.assertEqual(utils.join_words('-Office supplies'), 'Office supplies')

    def test_sanitize_args(self):
        self.assertDictEqual(utils.sanitize_fh_args({'_sort': ['-Office supplies']}),
                             {'_sort': ['Office supplies']})

    @unittest.skip('NER is unstable.')
    def test_ner(self):
        sent = nlp(
            """
            US President Donald Trump is an entrepreneur and
            used to run his own reality show named 'The Apprentice'."""
        )
        ents = utils.ner(sent, matcher)
        self.assertSetEqual(
            set([c.text for c in utils.unoverlap(ents)]),
            {
                "Donald Trump",
                "Apprentice",
                "US President",
                "President Donald",
                "entrepreneur",
                "reality show"
            },
        )

    def test_sanitize_indices(self):
        self.assertEqual(utils.sanitize_indices((3, 3), 0), 0)
        self.assertEqual(utils.sanitize_indices((3, 3), 1), 1)
        self.assertEqual(utils.sanitize_indices((3, 3), 2), -1)
        self.assertEqual(utils.sanitize_indices((3, 3), 0, 1), 0)
        self.assertEqual(utils.sanitize_indices((3, 3), 1, 1), 1)
        self.assertEqual(utils.sanitize_indices((3, 3), 2, 1), -1)


if __name__ == "__main__":
    unittest.main()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Templates used in Gramex NLG.
"""
import random
import re


from nlg import grammar as G  # noqa: N812

TEMPLATES = {
    "extreme": "{subject} {verb} the {adjective} {object}.",
    "comparison": "{subject} {verb} {quant} {adjective} than {object}.",
}

FUNC_EXPR = re.compile(r"^(?P<filter>[a-z|A-Z]{1}[a-z|A-Z|_|\d]+)\((?P<colname>.*)\)")


def parse_func_expr(expr):
    m = re.match(FUNC_EXPR, expr)
    if m is None:
        raise ValueError("Invalid Expression")
    return {"colname": m.group("colname"), "filter": m.group("filter")}


class Description(object):
    """Generate general purpose natural language descriptions of dataframes."""

    prefixes = ["This dataset contains "]
    metadata_tmpl = "{rowname} for {n_rows} {entities}."

    def __init__(self, df, entity="", attributes=""):
        """

        Parameters
        ----------
        df : `pandas.DataFrame`
            The source dataframe
        entity : str
            A named entity to represent the rows in the dataframe. For example,
            in Fisher's Iris dataset, the entity would be 'Iris flower
            samples', and in the Boston house pricing dataset it would be
            'Houses in Boston'.
        attributes : str
            A collective name to represent which attributes of `entity` are
            captured in the columns of the dataframe. For example, in the iris
            dataset, this would be 'petal and sepal measurements' and in the
            Boston house pricing dataset this would be 'real estate details'.
        """
        self.df = df
        self.attributes = attributes
        self.entity = entity
        # self.clean()
        self.categoricals = [
            c
            for c, d in df.dtypes.iteritems()
            if d.name in ("object", "bool", "category")
        ]
        self.numericals = [c for c, d in df.dtypes.iteritems() if d in (float, int)]
        self.desc = df[self.numericals].describe()
        self.indices = self.find_possible_indices()

    def find_possible_indices(self):
        """Find columns from the dataframe which can serve as indices.

        Any categorical column from the dataset which contains as many unique
        elements as the length of the dataframe can be used as an index."""
        indices = []
        for col in self.categoricals:
            if self.df[col].nunique() == self.df.shape[0]:
                indices.append(col)
        return indices

    def sentences(self):
        """Get the final narrative as a list of sentences."""
        return (
            self.get_metadata()
            + self.narrate_categoricals()  # NOQA: W503
            + self.narrate_numericals()  # NOQA: W503
        )

    def render(self, sep="\n", *args, **kwargs):
        """Generate the narration string.

        Parameters
        ----------
        sep : str
            Separator used to join individual sentences.

        Returns
        -------
        str
            The final narration.
        """
        return sep.join(self.sentences())

    def clean(self):
        self.df.dropna(inplace=True)
        self.df.drop_duplicates(inplace=True)

    def get_metadata(self):
        """Get the metadata of the dataframe as an English sentence. It
        contains details like the entity, attributes and the size of the
        dataset."""
        if self.indices and not self.entity:
            self.entity = random.choice(self.indices)
        entity = G.plural(self.entity)
        prefix = [random.choice(self.prefixes)]
        prefix.append(
            self.metadata_tmpl.format(
                rowname=self.attributes, n_rows=self.df.shape[0], entities=entity
            )
        )
        return prefix

    def common_categoricals(self, top_n=5):
        """Find the most common occurences of each categorical feature.

        Parameters
        ----------
        top_n : int
            Number of most frequent values to count.

        Returns
        -------
        dict
            A dictionary with keys as the column names and values as lists of
            the most common occurences of the feature.
        """
        results = {}
        if isinstance(top_n, int):
            top_n = [top_n] * len(self.categoricals)
        for i, colname in zip(top_n, self.categoricals):
            if colname not in self.indices:
                vcs = self.df[colname].value_counts()
                results[colname] = vcs[:i].index
        return results

    def narrate_categoricals(self, top_n=5):
        """Generate sentences that describe categorical variables.

        Parameters
        ----------
        top_n : int
            Number of most frequent values to count.

        Returns
        -------
        list
            A list of sentences.
        """
        sentences = []
        for k, v in self.common_categoricals(top_n).items():
            k = G.plural(k)
            n_items = len(v)
            values = G.concatenate_items(v)
            if n_items < top_n:
                top_n = n_items
                sent = "The {0} unique {1} are {2}."
            else:
                sent = "The top {0} {1} are {2}."
            sentences.append(sent.format(top_n, k, values))
        return sentences

    def narrate_numericals(self):
        """Generate sentences that describe numerical variables.

        Returns
        -------
        list
            A list of sentences.
        """
        sentences = []
        tmpl = "{colname} {verb} between {min} and {max} at an average of {mean}."
        for col in self.desc:
            colname = col.capitalize() if not col.isupper() else col
            verb = "vary" if G.is_plural(col) else "varies"
            sentences.append(
                tmpl.format(
                    colname=colname,
                    min=self.desc[col]["min"],
                    max=self.desc[col]["max"],
                    mean=self.desc[col]["mean"],
                    verb=verb,
                )
            )
        return sentences



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Tests for the nlg.grammar module.
"""
import unittest
import nlg.grammar as G  # noqa: N812


class TestGrammar(unittest.TestCase):

    def test_is_plural(self):
        self.assertTrue(G.is_plural_noun("languages"))
        # self.assertTrue(G.is_plural("geese"))
        self.assertTrue(G.is_plural_noun("bacteria"))
        self.assertTrue(G.is_plural_noun("Office supplies"))

    def test_concatenate_items(self):
        self.assertEqual(G.concatenate_items("abc"), "a, b and c")
        self.assertEqual(G.concatenate_items([1, 2, 3], sep=""), "123")
        self.assertFalse(G.concatenate_items([]))

    def test_pluralize(self):
        self.assertEqual(G.plural("language"), "languages")
        self.assertEqual(G.plural("languages"), "languages")
        self.assertEqual(G.plural("bacterium"), "bacteria")
        self.assertEqual(G.plural("goose"), "geese")

    def test_singular(self):
        self.assertEqual(G.singular("languages"), "language")
        self.assertEqual(G.singular("language"), "language")
        self.assertEqual(G.singular("bacteria"), "bacterium")
        # self.assertEqual(G.singular("geese"), "goose")

    def test_pluralize_by(self):
        self.assertEqual(G.pluralize_by("language", [1, 2]), "languages")
        self.assertEqual(G.pluralize_by("languages", [1]), "language")
        self.assertEqual(G.pluralize_by("language", []), "language")
        self.assertEqual(G.pluralize_by("language", 1), "language")
        self.assertEqual(G.pluralize_by("language", 2), "languages")



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
"""
NLG Setup.
"""

import builtins
from setuptools import setup, find_packages


builtins.__NLG_SETUP__ = True

# Setuptools config
NAME = "nlg"
DESCRIPTION = "Natural Language Generation framework for Python."
with open('README.rst', encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()
MAINTAINER = 'Jaidev Deshpande'
MAINTAINER_EMAIL = 'jaidev.deshpande@gramener.com'
URL = "https://github.com/gramener/gramex-nlg"
DOWNLOAD_URL = 'https://pypi.org/project/nlg/#files'
LICENSE = 'MIT'
PROJECT_URLS = {
    'Bug Tracker': 'https://github.com/gramener/gramex-nlg/issues',
    'Documentation': 'https://learn.gramener.com/guide/nlg',
    'Source Code': 'https://github.com/gramener/gramex-nlg'
}

# Requirements
install_requires = [
    'gramex',
    'humanize',
    'inflect',
    'spacy==2.1.8',
]

# Setup
import nlg  # NOQA: E402
setup(
    name=NAME,
    maintainer=MAINTAINER,
    maintainer_email=MAINTAINER_EMAIL,
    description=DESCRIPTION,
    license=LICENSE,
    url=URL,
    download_url=DOWNLOAD_URL,
    version=nlg.__version__,
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=install_requires
)



# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Miscellaneous utilities.
"""
import os.path as op
import re

from tornado.template import Template

from gramex.data import filter as grmfilter  # NOQA: F401

NP_RULES = {
    'NP1': [{'POS': 'PROPN', 'OP': '+'}],
    'NP2': [{'POS': 'NOUN', 'OP': '+'}],
    'NP3': [{'POS': 'ADV', 'OP': '+'}, {'POS': 'VERB', 'OP': '+'}],
    'NP4': [{'POS': 'ADJ', 'OP': '+'}, {'POS': 'VERB', 'OP': '+'}],
    'QUANT': [{'POS': 'NUM', 'OP': '+'}]
}

_spacy = {
    'model': False,
    'lemmatizer': False,
    'matcher': False
}


def _locate_app_config():
    return op.join(op.dirname(__file__), 'app', 'gramex.yaml')


def load_spacy_model():
    """Load the spacy model when required."""
    if not _spacy['model']:
        from spacy import load
        nlp = load('en_core_web_sm')
        _spacy['model'] = nlp
    else:
        nlp = _spacy['model']
    return nlp


def get_lemmatizer():
    if not _spacy['lemmatizer']:
        from spacy.lang.en import LEMMA_INDEX, LEMMA_EXC, LEMMA_RULES
        from spacy.lemmatizer import Lemmatizer
        lemmatizer = Lemmatizer(LEMMA_INDEX, LEMMA_EXC, LEMMA_RULES)
        _spacy['lemmatizer'] = lemmatizer
    else:
        lemmatizer = _spacy['lemmatizer']
    return lemmatizer


def make_np_matcher(nlp, rules=NP_RULES):
    """Make a rule based noun phrase matcher.

    Parameters
    ----------
    nlp : `spacy.lang`
        The spacy model to use.
    rules : dict, optional
        Mapping of rule IDS to spacy attribute patterns, such that each mapping
        defines a noun phrase structure.

    Returns
    -------
    `spacy.matcher.Matcher`
    """
    if not _spacy['matcher']:
        from spacy.matcher import Matcher
        matcher = Matcher(nlp.vocab)
        for k, v in rules.items():
            matcher.add(k, None, v)
        _spacy['matcher'] = matcher
    else:
        matcher = _spacy['matcher']
    return matcher


def render_search_result(text, results, **kwargs):
    for token, tokenlist in results.items():
        tmpl = [t for t in tokenlist if t.get('enabled', False)][0]
        text = text.replace(token, '{{{{ {} }}}}'.format(tmpl['tmpl']))
    return Template(text).generate(**kwargs).decode('utf-8')


def join_words(x, sep=' '):
    return sep.join(re.findall(r'\w+', x, re.IGNORECASE))


class set_nlg_gramopt(object):  # noqa: class to be used as a decorator
    """Decorator for adding callables to grammar options of the webapp.
    """
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self, func):
        func.gramopt = True
        for k, v in self.kwargs.items():
            if not getattr(func, k, False):
                setattr(func, k, v)
        return func


def is_overlap(x, y):
    """Whether the token x is contained within any span in the sequence y."""
    if 'NUM' in [c.pos_ for c in x]:
        return False
    return any([x.text in yy for yy in y])


def unoverlap(tokens):
    """From a set of tokens, remove all tokens that are contained within
    others."""
    textmap = {c.text: c for c in tokens}
    text_tokens = textmap.keys()
    newtokens = []
    for token in text_tokens:
        if not is_overlap(textmap[token], text_tokens - {token}):
            newtokens.append(token)
    return [textmap[t] for t in newtokens]


def ner(doc, matcher, match_ids=False, remove_overlap=True):
    """Find all NEs and other nouns in a spacy doc.

    Parameters
    ----------
    doc: spacy.tokens.doc.Doc
        The document in which to search for entities.
    matcher: spacy.matcher.Matcher
        The rule based matcher to use for finding noun phrases.
    match_ids: list, optional
        IDs from the spacy matcher to filter from the matches.
    remove_overlap: bool, optional
        Whether to remove overlapping tokens from the result.

    Returns
    -------
    list
        List of spacy.token.span.Span objects.
    """
    entities = set()
    for span in doc.ents:
        newtokens = [c for c in span if not c.is_space]
        if newtokens:
            newspan = doc[newtokens[0].i: (newtokens[-1].i + 1)]
            entities.add(newspan)
    if not match_ids:
        entities.update([doc[start:end] for _, start, end in matcher(doc)])
    else:
        for m_id, start, end in matcher(doc):
            if matcher.vocab.strings[m_id] in match_ids:
                entities.add(doc[start:end])
    if remove_overlap:
        entities = unoverlap(entities)
    return entities


def sanitize_indices(shape, i, axis=0):
    n = shape[axis]
    if i <= n // 2:
        return i
    return -(n - i)


def sanitize_text(text, d_round=2):
    """All text cleaning and standardization logic goes here."""
    nums = re.findall(r'\d+\.\d+', text)
    for num in nums:
        text = re.sub(num, str(round(float(num), d_round)), text)
    return text


def sanitize_df(df, d_round=2, **options):
    """All dataframe cleaning and standardizing logic goes here."""
    for c in df.columns[df.dtypes == float]:
        df[c] = df[c].round(d_round)
    return df


def sanitize_fh_args(args, func=join_words):
    for k, v in args.items():
        args[k] = [join_words(x) for x in v]
    return args


def add_html_styling(template, style):
    """Add HTML styling spans to template elements.

    Parameters
    ----------
    template : str
        A tornado template
    style : dict or bool
        If False, no styling is added.
        If True, a default bgcolor is added to template variables.
        If dict, expected to contain HTML span styling elements.

    Returns
    -------
    str
        Modified template with each variabled stylized.

    Example
    -------
    >>> t = 'Hello, {{ name }}!'
    >>> add_html_styling(t, True)
    'Hello, <span style='background-color:#c8f442'>{{ name }}</span>!'
    >>> add_html_styling(t, False)
    'Hello, {{ name }}!'
    >>> add_html_style(t, {'background-color': '#ffffff', 'font-family': 'monospace'})
    'Hello, <span style='background-color:#c8f442;font-family:monospace'>{{ name }}</span>!'
    """

    if not style:
        return template
    pattern = re.compile(r'\{\{[^\{\}]+\}\}')
    if isinstance(style, dict):
        # convert the style dict into a stylized HTML span
        spanstyle = ';'.join(['{}:{}'.format(k, v) for k, v in style.items()])
    else:
        spanstyle = 'background-color:#c8f442'
    for m in re.finditer(pattern, template):
        token = m.group()
        repl = '<span style="{ss}">{token}</span>'.format(
            ss=spanstyle, token=token)
        template = re.sub(re.escape(token), repl, template, 1)
    return '<p>{template}</p>'.format(template=template)



from inflect import engine
from tornado.template import Template

from nlg.utils import load_spacy_model, set_nlg_gramopt, get_lemmatizer

infl = engine()


def is_plural_noun(text):
    """Whether given text is a plural noun."""
    doc = load_spacy_model()(text)
    for t in list(doc)[::-1]:
        if not t.is_punct:
            return t.tag_ in ('NNS', 'NNPS')
    return False


is_singular_noun = lambda x: not is_plural_noun(x)  # NOQA: E731


@set_nlg_gramopt(source='G', fe_name='Concate Items')
def concatenate_items(items, sep=', '):
    """Concatenate a sequence of tokens into an English string.

    Parameters
    ----------

    items : list-like
        List / sequence of items to be printed.
    sep : str, optional
        Separator to use when generating the string

    Returns
    -------
    str
    """
    if len(items) == 0:
        return ""
    if len(items) == 1:
        return items[0]
    items = list(map(str, items))
    if sep == ', ':
        s = sep.join(items[:-1])
        s += ' and ' + items[-1]
    else:
        s = sep.join(items)
    return s


@set_nlg_gramopt(source='G', fe_name='Pluralize')
def plural(word):
    """Pluralize a word.

    Parameters
    ----------

    word : str
        word to pluralize

    Returns
    -------
    str
        Plural of `word`
    """
    if not is_plural_noun(word):
        word = infl.plural(word)
    return word


@set_nlg_gramopt(source='G', fe_name='Singularize')
def singular(word):
    """
    Singularize a word.

    Parameters
    ----------
    word : str
        Word to singularize.

    Returns
    -------
    str
        Singular of `word`.
    """
    if is_plural_noun(word):
        word = infl.singular_noun(word)
    return word


# @set_nlg_gramopt(source='G', fe_name='Pluralize by')
def pluralize_by(word, by):
    """
    Pluralize a word depending on another argument.

    Parameters
    ----------
    word : str
        Word to pluralize
    by : any
        Any object checked for a pluralish value. If a sequence, it must have
        length greater than 1 to qualify as plural.

    Returns
    -------
    str
        Plural or singular of `word`.
    """
    if hasattr(by, '__iter__'):
        if len(by) > 1:
            word = plural(word)
        else:
            word = singular(word)
    else:
        if by > 1:
            word = plural(word)
        else:
            word = singular(word)
    return word


# @set_nlg_gramopt(source='G', fe_name='Pluralize like')
def pluralize_like(x, y):
    """
    Pluralize a word if another is a plural.

    Parameters
    ----------
    x : str
        The word to pluralize.
    y : str
        The word to check.

    Returns
    -------
    str
        Plural of `x` if `y` is plural, else singular.
    """
    if not is_plural_noun(y):
        return singular(x)
    return plural(x)


@set_nlg_gramopt(source='str', fe_name='Capitalize')
def capitalize(word):
    return word.capitalize()


@set_nlg_gramopt(source='str', fe_name='Lowercase')
def lower(word):
    return word.lower()


@set_nlg_gramopt(source='str', fe_name='Swapcase')
def swapcase(word):
    return word.swapcase()


@set_nlg_gramopt(source='str', fe_name='Title')
def title(word):
    return word.title()


@set_nlg_gramopt(source='str', fe_name='Uppercase')
def upper(word):
    return word.upper()


# @set_nlg_gramopt(source='G', fe_name='Lemmatize')
def lemmatize(word, target_pos):
    return get_lemmatizer()(word, target_pos)


def _token_inflections(x, y):
    """
    If two words share the same root, find lexical changes required for turning
    one into another.

    Parameters
    ----------
    x : spacy.token.Tokens
    y : spacy.token.Tokens

    Examples
    --------
    >>> _token_inflections('language', 'Language')
    'upper'
    >>> _token_inflections('language', 'languages')
    'plural'
    """
    if x.lemma_ != y.lemma_:
        return False
    if len(x.text) == len(y.text):
        for methname in ['capitalize', 'lower', 'swapcase', 'title', 'upper']:
            func = lambda x: getattr(x, methname)()  # NOQA: E731
            if func(x.text) == y.text:
                return globals()[methname]
    # check if x and y are singulars or plurals of each other.
    if is_singular_noun(y.text):
        if singular(x.text).lower() == y.text.lower():
            return singular
    elif is_plural_noun(y.text):
        if plural(x.text).lower() == y.text.lower():
            return plural
    # Disable detecting inflections until they can be
    # processed without intervention.
    # if x.pos_ != y.pos_:
    #     return lemmatize
    return False


def find_inflections(text, search, fh_args, df):
    """
    Find lexical inflections between words in input text and the search results
    obtained from FormHandler arguments and dataframes.

    Parameters
    ----------
    text : str
        Input text
    search : nlg.search.DFSearchResults
        The DFSearchResults object corresponding to `text` and `df`
    fh_args : dict
        FormHandler arguments.
    df : pandas.DataFrame
        The source dataframe.

    Returns
    -------
    dict
        With keys as tokens found in the dataframe or FH args, and values as
        list of inflections applied on them to make them closer match tokens in `text`.
    """
    nlp = load_spacy_model()
    text = nlp(text)
    inflections = {}
    for token, tklist in search.items():
        tmpl = [t['tmpl'] for t in tklist if t.get('enabled', False)][0]
        rendered = Template('{{{{ {} }}}}'.format(tmpl)).generate(
            df=df, fh_args=fh_args).decode('utf8')
        if rendered != token:
            x = nlp(rendered)[0]
            y = text[[c.text for c in text].index(token)]
            infl = _token_inflections(x, y)
            if infl:
                inflections[token] = [infl]
    return inflections
